
// Operations on stepped block distributed ranges.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"

inline int max(int x, int y) {return x > y ? x : y ; }
inline int min(int x, int y) {return x < y ? x : y ; }


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `StepRangeRep': stepped block-distributed template range.     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void StepRangeRep :: location(Location* loc, const int glb) const {

  if(glb < stepSub) {
    loc->sub = glb % (blockSize + 1) ;
    loc->crd = glb / (blockSize + 1) ;
  }
  else {
    loc->sub = (glb - stepCrd) % blockSize ;
    loc->crd = (glb - stepCrd) / blockSize ;
  }
  loc->dim = dimension ;
}

RangeRep* StepRangeRep :: shell() const {
  return new CollapsedRangeRep(blockSize) ;
}

void StepRangeRep :: block(Block* blk, const int ker_glb) const {

  // `ker_glb' is a global subscript in the kernel range.

  blk->sub_stp = 1 ;
  blk->glb_stp = 1 ;

  blk->sub_bas = 0 ;
  if(ker_glb < stepCrd) {
    blk->glb_bas = (blockSize + 1) * ker_glb ;

    blk->count   = blockSize + 1 ;
  }
  else {
    blk->glb_bas = blockSize * ker_glb + stepCrd ;

    blk->count   = blockSize ;
  }
}

void StepRangeRep :: block2(Block* blk, const int ker_glb,
                            const int count, const int base) const {
  blk->sub_stp = 1 ;
  blk->glb_stp = 1 ;

  int blockStart, blockStartOffset, blockFinishOffset ;
  if(ker_glb < stepCrd) {
    blockStart        = ker_glb * (blockSize + 1) ;
    blockStartOffset  = blockStart - base ;
    blockFinishOffset = blockStartOffset + blockSize ;
  }
  else {
    blockStart        = ker_glb * blockSize + stepCrd;
    blockStartOffset  = blockStart - base ;
    blockFinishOffset = blockStartOffset + blockSize - 1 ;
  }

  int i_lo, i_hi ;

  i_lo = max(blockStartOffset, 0) ;
  i_hi = min(blockFinishOffset, count - 1) ;

  blk->sub_bas = base + i_lo - blockStart ;
  blk->glb_bas = i_lo ;

  blk->count   = i_hi - i_lo + 1 ;
}

void StepRangeRep :: block3(Block* blk, const int ker_glb,
                            const int count, const int base,
                            const int stride) const {
  blk->sub_stp = stride ;
  blk->glb_stp = 1 ;

  int blockStart, blockStartOffset, blockFinishOffset ;
  if(ker_glb < stepCrd) {
    blockStart        = ker_glb * (blockSize + 1) ;
    blockStartOffset  = blockStart - base ;
    blockFinishOffset = blockStartOffset + blockSize ;
  }
  else {
    blockStart        = ker_glb * blockSize + stepCrd;
    blockStartOffset  = blockStart - base ;
    blockFinishOffset = blockStartOffset + blockSize - 1 ;
  }

  int i_lo, i_hi ;

  if(stride > 0) {
    i_lo = max((blockStartOffset + stride - 1) / stride, 0) ;
    i_hi = min(blockFinishOffset / stride, count - 1) ;
  }
  else {
    i_lo = max((blockFinishOffset + stride + 1) / stride, 0) ;
    i_hi = min(blockStartOffset / stride, count - 1) ;
  }

  blk->sub_bas = base + stride * i_lo - blockStart ;
  blk->glb_bas = i_lo ;

  blk->count   = i_hi - i_lo + 1 ;
}

void StepRangeRep :: crds2(int* ker_count, int* ker_bas, int* ker_str,
                           const int count, const int base) const {

  const int ker_top = to_ker_glb(base + count - 1) ;

  *ker_bas = to_ker_glb(base) ;

  *ker_count = extent ? ker_top - *ker_bas + 1 : 0 ;
  *ker_str   = 1 ;
}

void StepRangeRep :: crds3(int* ker_count, int* ker_bas, int* ker_str,
                           const int count, const int base,
                           const int stride) const {

  const int ker_top = to_ker_glb(base + stride * (count - 1)) ;

  *ker_bas = to_ker_glb(base) ;

  if(stride > 0) {
    *ker_count = extent ? ker_top - *ker_bas + 1 : 0 ;
    *ker_str   = 1 ;
  }
  else {
    *ker_count = extent ? *ker_bas - ker_top + 1 : 0 ;
    *ker_str   = -1 ;
  }
}

int StepRangeRep :: volume() const {
  return stepCrd ? blockSize + 1 : blockSize ;
}

int StepRangeRep :: glb(const int sub, const int ker_glb) const {
  if(ker_glb < stepCrd)
    return (blockSize + 1) * ker_glb + sub ;
  else
    return blockSize * ker_glb + stepCrd + sub ;
}

int StepRangeRep :: isAligned(RangeRep* x) const {
  if(x == (RangeRep*) this) return 1 ;
  if(x->format() == DIST_STEP && x->dim() == dimension) {
    StepRangeRep* t = (StepRangeRep*) x->tem() ;
    return t->blockSize == blockSize && t->stepCrd == stepCrd &&
           x->size() == extent && x->str() == 1 ;
  }
  else
    return 0 ;
}

void StepRangeRep :: show() {
  AdlibPrintf("StepRangeRep :: show: size = %d ker = %p lev = %d\n",
              extent, kernel, level) ; 
}

MEMORY_POOL_DEFN(StepRangeRep)


// -------------------------------------------------------------
// Things to do
//

